import 'dart:async';

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:fireflow/fireflow.dart';
import 'package:flutter/material.dart';
import 'package:firebase_auth/firebase_auth.dart';
import 'package:flutterflow_paginate_firestore/paginate_firestore.dart';

/// ChatRoomMessageList is a widget that displays a list of messages in a chat room.
///
class ChatRoomMessageList extends StatefulWidget {
  const ChatRoomMessageList({
    Key? key,
    this.width,
    this.height,
    this.otherUserPublicDataDocument,
    this.chatRoomDocumentReference,
    required this.onMyMessage,
    required this.onOtherMessage,
    required this.onEmpty,
  })  : assert(
            (otherUserPublicDataDocument != null &&
                    chatRoomDocumentReference == null) ||
                (otherUserPublicDataDocument == null &&
                    chatRoomDocumentReference != null),
            "You must set only one of otherUserPublicDataDocument or chatRoomDocumentReference."),
        super(key: key);

  final double? width;
  final double? height;
  final DocumentReference? otherUserPublicDataDocument;
  final DocumentReference? chatRoomDocumentReference;

  final Widget Function(ChatRoomMessageModel) onMyMessage;
  final Widget Function(ChatRoomMessageModel) onOtherMessage;
  final Widget onEmpty;

  @override
  _ChatRoomMessageListState createState() => _ChatRoomMessageListState();
}

class _ChatRoomMessageListState extends State<ChatRoomMessageList> {
  User get my => UserService.instance.my;
  DocumentReference get myReference => UserService.instance.ref;

  late final StreamSubscription subscriptionNewMessage;

  bool ensureChatRoomExists = false;

  /// The chat room document references of the 1:1 chat room between you and the other user.
  List<DocumentReference> get youAndMeRef => [
        UserService.instance.doc(widget.otherUserPublicDataDocument!.id),
        myReference,
      ];

  /// The chat room document reference.
  DocumentReference get chatRoomRef {
    if (isGroupChat) {
      return widget.chatRoomDocumentReference!;
    } else {
      return ChatService.instance.room(
          ([my.uid, widget.otherUserPublicDataDocument!.id]..sort()).join('-'));
    }
  }

  bool get isGroupChat => widget.chatRoomDocumentReference != null;
  bool get isSingleChat => !isGroupChat;

  @override
  void initState() {
    super.initState();

    init();
  }

  /// Initialize the chat room.
  ///
  init() async {
    // Set the chat room ref where I am chat in
    AppService.instance.currentChatRoomReference = chatRoomRef;

    if (isSingleChat) {
      // For 1:1 chat,
      //  - create a chat room if it does not exist.
      //  - just make the last message seen by you.
      await chatRoomRef.set({
        'userDocumentReferences': youAndMeRef,
        'lastMessageSeenBy': FieldValue.arrayUnion([myReference]),
      }, SetOptions(merge: true));
    } else {
      // For group chat,
      //  - users can only invited by other user.
      //  - just make the last message seen by you.
      await chatRoomRef.update({
        'lastMessageSeenBy': FieldValue.arrayUnion([myReference]),
      });
    }

    /// When the chat begins, the app try to get messages before the chat room exists.
    /// This may lead permission error and showing an infinite loader.
    /// Rebuiding the screen helps to avoid the infinite loader.
    setState(() {
      ensureChatRoomExists = true;
    });

    // Listen for new message, and make it read by you.
    subscriptionNewMessage = chatRoomRef.snapshots().listen((snapshot) {
      final room = ChatRoomModel.fromSnapshot(snapshot);

      /// If the signed-in user have not seen the message, then make it seen.
      if (room.lastMessageSeenBy.contains(myReference) == false) {
        chatRoomRef.update({
          'lastMessageSeenBy': FieldValue.arrayUnion([myReference])
        });
      }
    });
  }

  @override
  void dispose() {
    AppService.instance.currentChatRoomReference = null;
    subscriptionNewMessage.cancel();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    if (ensureChatRoomExists == false) return const SizedBox.shrink();
    return PaginateFirestore(
      reverse: true,
      // item builder type is compulsory.
      itemBuilder: (context, documentSnapshots, index) {
        final message =
            ChatRoomMessageModel.fromSnapshot(documentSnapshots[index]);

        if (message.isMine) {
          return widget.onMyMessage(message);
        } else {
          return widget.onOtherMessage(message);
        }
      },

      /// Get messages of the chat room.
      query: FirebaseFirestore.instance
          .collection('chat_room_messages')
          .where('chatRoomDocumentReference', isEqualTo: chatRoomRef)
          .orderBy('sentAt', descending: true),
      //Change types accordingly
      itemBuilderType: PaginateBuilderType.listView,
      // to fetch real-time data
      isLive: true,
      onEmpty: widget.onEmpty,
    );
  }
}
